"""
<Module Name>
  handlers.py

<Author>
  Santiago Torres-Arias <santiago@nyu.edu>

<Started>
  Jan 15, 2020

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides links from signatures/algorithms to modules implementing
  the signature verification and key parsing.
"""

from securesystemslib.gpg import rsa
from securesystemslib.gpg import dsa
from securesystemslib.gpg import eddsa

# See section 9.1. (public-key algorithms) of RFC4880 (-bis8)
SUPPORTED_SIGNATURE_ALGORITHMS = {
    0x01: {
      "type":"rsa",
      "method": "pgp+rsa-pkcsv1.5",
      "handler": rsa
    },
    0x11: {
      "type": "dsa",
      "method": "pgp+dsa-fips-180-2",
      "handler": dsa
    },
    0x16: {
      "type": "eddsa",
      "method": "pgp+eddsa-ed25519",
      "handler": eddsa
    }
}

SIGNATURE_HANDLERS = {
    "rsa": rsa,
    "dsa": dsa,
    "eddsa": eddsa
}



"""
<Module Name>
  gpg

<Author>
  Santiago Torres-Arias <santiago@nyu.edu>

<Started>
  Nov 15, 2017

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  This module was written due to the lack of other python (such as pygpg)
  modules that can provide an abstraction to the RFC4480 encoded messages from
  GPG. The closest candidate we could find was the python bindings for gpgme,
  we oped to use a Popen-based python-only construction given that gpgme is
  often shipped separately and other popular tools using gpg (e.g., git) don't
  use these bindings either. This is because users willing to use gpg signing
  are almost guaranteed to have gpg installed, yet the same assumption can't be
  made for the gpgme python bindings.
"""



"""
<Module Name>
  common.py

<Author>
  Santiago Torres-Arias <santiago@nyu.edu>

<Started>
  Nov 15, 2017

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides algorithm-agnostic gpg public key and signature parsing functions.
  The functions select the appropriate functions for each algorithm and
  call them.

"""
import struct
import binascii
import logging
import collections

from securesystemslib import formats
from securesystemslib.gpg import util as gpg_util
from securesystemslib.gpg.exceptions import (PacketVersionNotSupportedError,
    SignatureAlgorithmNotSupportedError, KeyNotFoundError, PacketParsingError)
from securesystemslib.gpg.constants import (
    PACKET_TYPE_PRIMARY_KEY, PACKET_TYPE_USER_ID, PACKET_TYPE_USER_ATTR,
    PACKET_TYPE_SUB_KEY, PACKET_TYPE_SIGNATURE,
    SUPPORTED_PUBKEY_PACKET_VERSIONS, SIGNATURE_TYPE_BINARY,
    SIGNATURE_TYPE_CERTIFICATES, SIGNATURE_TYPE_SUB_KEY_BINDING,
    SUPPORTED_SIGNATURE_PACKET_VERSIONS,
    FULL_KEYID_SUBPACKET, PARTIAL_KEYID_SUBPACKET,
    SHA1,SHA256, SHA512, KEY_EXPIRATION_SUBPACKET, PRIMARY_USERID_SUBPACKET,
    SIG_CREATION_SUBPACKET)
from securesystemslib.gpg.handlers import (
    SIGNATURE_HANDLERS, SUPPORTED_SIGNATURE_ALGORITHMS)

log = logging.getLogger(__name__)


def parse_pubkey_payload(data):
  """
  <Purpose>
    Parse the passed public-key packet (payload only) and construct a
    public key dictionary.

  <Arguments>
    data:
          An RFC4880 public key packet payload as described in section 5.5.2.
          (version 4) of the RFC.

          NOTE: The payload can be parsed from a full key packet (header +
          payload) by using securesystemslib.gpg.util.parse_packet_header.

          WARNING: this doesn't support armored pubkey packets, so use with
          care. pubkey packets are a little bit more complicated than the
          signature ones

  <Exceptions>
    ValueError
          If the passed public key data is empty.

    securesystemslib.gpg.exceptions.PacketVersionNotSupportedError
          If the packet version does not match
          securesystemslib.gpg.constants.SUPPORTED_PUBKEY_PACKET_VERSIONS

    securesystemslib.gpg.exceptions.SignatureAlgorithmNotSupportedError
          If the signature algorithm does not match one of
          securesystemslib.gpg.constants.SUPPORTED_SIGNATURE_ALGORITHMS

  <Side Effects>
    None.

  <Returns>
    A public key in the format securesystemslib.formats.GPG_PUBKEY_SCHEMA

  """
  if not data:
    raise ValueError("Could not parse empty pubkey payload.")

  ptr = 0
  keyinfo = {}
  version_number = data[ptr]
  ptr += 1
  if version_number not in SUPPORTED_PUBKEY_PACKET_VERSIONS:
    raise PacketVersionNotSupportedError(
        "Pubkey packet version '{}' not supported, must be one of {}".format(
        version_number, SUPPORTED_PUBKEY_PACKET_VERSIONS))

  # NOTE: Uncomment this line to decode the time of creation
  time_of_creation = struct.unpack(">I", data[ptr:ptr + 4])
  ptr += 4

  algorithm = data[ptr]

  ptr += 1

  # TODO: Should we only export keys with signing capabilities?
  # Section 5.5.2 of RFC4880 describes a public-key algorithm octet with one
  # of the values described in section 9.1 that could be used to determine the
  # capabilities. However, in case of RSA subkeys this field doesn't seem to
  # correctly encode the capabilities. It always has the value 1, i.e.
  # RSA (Encrypt or Sign).
  # For RSA public keys we would have to parse the subkey's signature created
  # with the master key, for the signature's key flags subpacket, identified
  # by the value 27 (see section 5.2.3.1.) containing a list of binary flags
  # as described in section 5.2.3.21.
  if algorithm not in SUPPORTED_SIGNATURE_ALGORITHMS:
    raise SignatureAlgorithmNotSupportedError("Signature algorithm '{}' not "
        "supported, please verify that your gpg configuration is creating "
        "either DSA, RSA, or EdDSA signatures (see RFC4880 9.1. Public-Key "
        "Algorithms).".format(algorithm))

  keyinfo['type'] = SUPPORTED_SIGNATURE_ALGORITHMS[algorithm]['type']
  keyinfo['method'] = SUPPORTED_SIGNATURE_ALGORITHMS[algorithm]['method']
  handler = SIGNATURE_HANDLERS[keyinfo['type']]
  keyinfo['keyid'] = gpg_util.compute_keyid(data)
  key_params = handler.get_pubkey_params(data[ptr:])

  return {
    "method": keyinfo['method'],
    "type": keyinfo['type'],
    "hashes": [formats.GPG_HASH_ALGORITHM_STRING],
    "creation_time": time_of_creation[0],
    "keyid": keyinfo['keyid'],
    "keyval" : {
      "private": "",
      "public": key_params
      }
    }


def parse_pubkey_bundle(data):
  """
  <Purpose>
    Parse packets from passed gpg public key data, associating self-signatures
    with the packets they correspond to, based on the structure of V4 keys
    defined in RFC4880 12.1 Key Structures.

    The returned raw key bundle may be used to further enrich the master key,
    with certified information (e.g. key expiration date) taken from
    self-signatures, and/or to verify that the parsed subkeys are bound to the
    primary key via signatures.

  <Arguments>
    data:
          Public key data as written to stdout by GPG_EXPORT_PUBKEY_COMMAND.

  <Exceptions>
    securesystemslib.gpg.exceptions.PacketParsingError
          If data is empty.
          If data cannot be parsed.

  <Side Effects>
    None.

  <Returns>
    A raw public key bundle where self-signatures are associated with their
    corresponding packets. See `key_bundle` for details.

  """
  if not data:
    raise PacketParsingError("Cannot parse keys from empty gpg data.")

  # Temporary data structure to hold parsed gpg packets
  key_bundle = {
    PACKET_TYPE_PRIMARY_KEY: {
      "key": {},
      "packet": None,
      "signatures": []
    },
    PACKET_TYPE_USER_ID: collections.OrderedDict(),
    PACKET_TYPE_USER_ATTR: collections.OrderedDict(),
    PACKET_TYPE_SUB_KEY: collections.OrderedDict()
  }

  # Iterate over gpg data and parse out packets of different types
  position = 0
  while position < len(data):
    try:
      packet_type, header_len, body_len, packet_length = \
          gpg_util.parse_packet_header(data[position:])

      packet = data[position:position+packet_length]
      payload = packet[header_len:]
      # The first (and only the first) packet in the bundle must be the master
      # key.  See RFC4880 12.1 Key Structures, V4 version keys
      # TODO: Do we need additional key structure assertions? e.g.
      # - there must be least one User ID packet, or
      # - order and type of signatures, or
      # - disallow duplicate packets
      if packet_type != PACKET_TYPE_PRIMARY_KEY and \
          not key_bundle[PACKET_TYPE_PRIMARY_KEY]["key"]:
        raise PacketParsingError("First packet must be a primary key ('{}'), "
            "got '{}'.".format(PACKET_TYPE_PRIMARY_KEY, packet_type))

      elif packet_type == PACKET_TYPE_PRIMARY_KEY and \
          key_bundle[PACKET_TYPE_PRIMARY_KEY]["key"]:
        raise PacketParsingError("Unexpected primary key.")

      # Fully parse master key to fail early, e.g. if key is malformed
      # or not supported, but also retain original packet for subkey binding
      # signature verification
      elif packet_type == PACKET_TYPE_PRIMARY_KEY:
        key_bundle[PACKET_TYPE_PRIMARY_KEY] = {
          "key": parse_pubkey_payload(bytearray(payload)),
          "packet": packet,
          "signatures": []
        }

      # Other non-signature packets in the key bundle include User IDs and User
      # Attributes, required to verify primary key certificates, and subkey
      # packets. For each packet we create a new ordered dictionary entry. We
      # use a dictionary to aggregate signatures by packet below,
      # and it must be ordered because each signature packet belongs to the
      # most recently parsed packet of a type.
      elif packet_type in {PACKET_TYPE_USER_ID, PACKET_TYPE_USER_ATTR,
          PACKET_TYPE_SUB_KEY}:
        key_bundle[packet_type][packet] = {
          "header_len": header_len,
          "body_len": body_len,
          "signatures": []
        }

      # The remaining relevant packets are signatures, required to bind subkeys
      # to the primary key, or to gather additional information about the
      # primary key, e.g. expiration date.
      # A signature corresponds to the most recently parsed packet of a type,
      # where the type is given by the availability of respective packets.
      # We test availability and assign accordingly as per the order of packet
      # types defined in RFC4880 12.1 (bottom-up).
      elif packet_type == PACKET_TYPE_SIGNATURE:
        for _type in [PACKET_TYPE_SUB_KEY, PACKET_TYPE_USER_ATTR,
            PACKET_TYPE_USER_ID]:
          if key_bundle[_type]:
            # Add to most recently added packet's signatures of matching type
            key_bundle[_type][next(reversed(key_bundle[_type]))]\
                ["signatures"].append(packet)
            break

        else:
          # If no packets are available for any of above types (yet), the
          # signature belongs to the primary key
          key_bundle[PACKET_TYPE_PRIMARY_KEY]["signatures"].append(packet)

      else:
        log.info("Ignoring gpg key packet '{}', we only handle packets of "
            "types '{}' (see RFC4880 4.3. Packet Tags).".format(packet_type,
            [PACKET_TYPE_PRIMARY_KEY, PACKET_TYPE_USER_ID,
            PACKET_TYPE_USER_ATTR, PACKET_TYPE_SUB_KEY,
            PACKET_TYPE_SIGNATURE]))

    # Both errors might be raised in parse_packet_header and in this loop
    except (PacketParsingError, IndexError) as e:
      raise PacketParsingError("Invalid public key data at position {}: {}."
          .format(position, e))

    # Go to next packet
    position += packet_length

  return key_bundle


def _assign_certified_key_info(bundle):
  """
  <Purpose>
    Helper function to verify User ID certificates corresponding to a gpg
    master key, in order to enrich the master key with additional information
    (e.g. expiration dates). The enriched master key is returned.

    NOTE: Currently we only consider User ID certificates. We can do the same
    for User Attribute certificates by iterating over
    bundle[PACKET_TYPE_USER_ATTR] instead of bundle[PACKET_TYPE_USER_ID], and
    replacing the signed_content constant '\xb4'  with '\xd1' (see RFC4880
    section 5.2.4. paragraph 4).

  <Arguments>
    bundle:
          GPG key bundle as parsed in parse_pubkey_bundle().

  <Exceptions>
    None.

  <Side Effects>
    None.

  <Returns>
    A public key in the format securesystemslib.formats.GPG_PUBKEY_SCHEMA.

  """
  # Create handler shortcut
  handler = SIGNATURE_HANDLERS[bundle[PACKET_TYPE_PRIMARY_KEY]["key"]["type"]]

  is_primary_user = False
  validity_period = None
  sig_creation_time = None

  # Verify User ID signatures to gather information about primary key
  # (see Notes about certification signatures in RFC 4880 5.2.3.3.)
  for user_id_packet, packet_data in bundle[PACKET_TYPE_USER_ID].items():
    # Construct signed content (see RFC4880 section 5.2.4. paragraph 4)
    signed_content = (bundle[PACKET_TYPE_PRIMARY_KEY]["packet"] +
        b"\xb4\x00\x00\x00" + user_id_packet[1:])
    for signature_packet in packet_data["signatures"]:
      try:
        signature = parse_signature_packet(signature_packet,
            supported_hash_algorithms={SHA1, SHA256, SHA512},
            supported_signature_types=SIGNATURE_TYPE_CERTIFICATES,
            include_info=True)
        # verify_signature requires a "keyid" even if it is short.
        # (see parse_signature_packet for more information about keyids)
        signature["keyid"] = signature["keyid"] or signature["short_keyid"]

      # TODO: Revise exception taxonomy:
      # It's okay to ignore some exceptions (unsupported algorithms etc.) but
      # we should blow up if a signature is malformed (missing subpackets).
      except Exception as e:
        log.info(e)
        continue

      if not bundle[PACKET_TYPE_PRIMARY_KEY]["key"]["keyid"].endswith(
          signature["keyid"]):
        log.info("Ignoring User ID certificate issued by '{}'.".format(
            signature["keyid"]))
        continue

      is_valid = handler.verify_signature(signature,
          bundle[PACKET_TYPE_PRIMARY_KEY]["key"], signed_content,
          signature["info"]["hash_algorithm"])

      if not is_valid:
        log.info("Ignoring invalid User ID self-certificate issued "
            "by '{}'.".format(signature["keyid"]))
        continue

      # If the signature is valid, we try to extract subpackets relevant to
      # the primary key, i.e. expiration time.
      # NOTE: There might be multiple User IDs per primary key and multiple
      # certificates per User ID. RFC4880 5.2.3.19. and last paragraph of
      # 5.2.3.3. provides some suggestions about ambiguity, but delegates the
      # responsibility to the implementer.

      # Ambiguity resolution scheme:
      # We take the key expiration time from the most recent certificate, i.e.
      # the certificate with the highest signature creation time. Additionally,
      # we prioritize certificates with primary user id flag set True. Note
      # that, if the ultimately prioritized certificate does not have a key
      # expiration time subpacket, we don't assign one, even if there were
      # certificates of lower priority carrying that subpacket.
      tmp_validity_period = \
          signature["info"]["subpackets"].get(KEY_EXPIRATION_SUBPACKET)

      # No key expiration time, go to next certificate
      if tmp_validity_period is None:
        continue

      # Create shortcut to mandatory pre-parsed creation time subpacket
      tmp_sig_creation_time = signature["info"]["creation_time"]

      tmp_is_primary_user = \
          signature["info"]["subpackets"].get(PRIMARY_USERID_SUBPACKET)

      if tmp_is_primary_user is not None:
        tmp_is_primary_user = bool(tmp_is_primary_user[0])

      # If we already have a primary user certified expiration date and this
      # is none, we don't consider it, and go to next certificate
      if is_primary_user and not tmp_is_primary_user:
        continue

      if not sig_creation_time or sig_creation_time < tmp_sig_creation_time:
        # This is the most recent certificate that has a validity_period and
        # doesn't have lower priority in regard to the primary user id flag. We
        # accept it the keys validty_period, until we get a newer value from
        # a certificate with higher priority.
        validity_period = struct.unpack(">I", tmp_validity_period)[0]
        # We also keep track of the used certificate's primary user id flag and
        # the signature creation time, for prioritization.
        is_primary_user = tmp_is_primary_user
        sig_creation_time = tmp_sig_creation_time

  if validity_period is not None:
    bundle[PACKET_TYPE_PRIMARY_KEY]["key"]["validity_period"] = validity_period

  return bundle[PACKET_TYPE_PRIMARY_KEY]["key"]


def _get_verified_subkeys(bundle):
  """
  <Purpose>
    Helper function to verify the subkey binding signature for all subkeys in
    the passed bundle in order to enrich subkeys with additional information
    (e.g. expiration dates). Only valid (i.e. parsable) subkeys that are
    verifiably bound to the the master key of the bundle are returned. All
    other subkeys are discarded.

  <Arguments>
    bundle:
          GPG key bundle as parsed in parse_pubkey_bundle().

  <Exceptions>
    None.

  <Side Effects>
    None.

  <Returns>
    A dictionary of public keys in the format
    securesystemslib.formats.GPG_PUBKEY_SCHEMA, with keyids as dict keys.

  """
  # Create handler shortcut
  handler = SIGNATURE_HANDLERS[bundle[PACKET_TYPE_PRIMARY_KEY]["key"]["type"]]

  # Verify subkey binding signatures and only keep verified keys
  # See notes about subkey binding signature in RFC4880 5.2.3.3
  verified_subkeys = {}
  for subkey_packet, packet_data in bundle[PACKET_TYPE_SUB_KEY].items():
    try:
      # Parse subkey if possible and skip if invalid (e.g. not-supported)
      subkey = parse_pubkey_payload(
          bytearray(subkey_packet[-packet_data["body_len"]:]))

    # TODO: Revise exception taxonomy
    except Exception as e:
      log.info(e)
      continue

    # Construct signed content (see RFC4880 section 5.2.4. paragraph 3)
    signed_content = (bundle[PACKET_TYPE_PRIMARY_KEY]["packet"] + b"\x99" +
        subkey_packet[1:])

    # Filter sub key binding signature from other signatures, e.g. subkey
    # binding revocation signatures
    key_binding_signatures = []
    for signature_packet in packet_data["signatures"]:
      try:
        signature = parse_signature_packet(signature_packet,
            supported_hash_algorithms={SHA1, SHA256, SHA512},
            supported_signature_types={SIGNATURE_TYPE_SUB_KEY_BINDING},
            include_info=True)
        # verify_signature requires a "keyid" even if it is short.
        # (see parse_signature_packet for more information about keyids)
        signature["keyid"] = signature["keyid"] or signature["short_keyid"]
        key_binding_signatures.append(signature)

      # TODO: Revise exception taxonomy
      except Exception as e:
        log.info(e)
        continue
    # NOTE: As per the V4 key structure diagram in RFC4880 section 12.1., a
    # subkey must be followed by exactly one Primary-Key-Binding-Signature.
    # Based on inspection of real-world keys and other parts of the RFC (e.g.
    # the paragraph below the diagram and paragraph 0x18: Subkey Binding
    # Signature in section 5.2.1.) the mandated signature is actually a
    # *subkey binding signature*, which in case of a signing subkey, must have
    # an *embedded primary key binding signature*.
    if len(key_binding_signatures) != 1:
      log.info("Ignoring subkey '{}' due to wrong amount of key binding "
          "signatures ({}), must be exactly 1.".format(subkey["keyid"],
          len(key_binding_signatures)))
      continue
    is_valid = handler.verify_signature(signature,
        bundle[PACKET_TYPE_PRIMARY_KEY]["key"], signed_content,
        signature["info"]["hash_algorithm"])

    if not is_valid:
      log.info("Ignoring subkey '{}' due to invalid key binding signature."
          .format(subkey["keyid"]))
      continue

    # If the signature is valid, we may also extract relevant information from
    # its "info" field (e.g. subkey expiration date) and assign to it to the
    # subkey here
    validity_period = \
        signature["info"]["subpackets"].get(KEY_EXPIRATION_SUBPACKET)
    if validity_period is not None:
      subkey["validity_period"] = struct.unpack(">I", validity_period)[0]

    verified_subkeys[subkey["keyid"]] = subkey

  return verified_subkeys


def get_pubkey_bundle(data, keyid):
  """
  <Purpose>
    Call function to extract and verify master key and subkeys from the passed
    gpg key data, where either the master key or one of the subkeys matches the
    passed keyid.

    NOTE:
    - If the keyid matches one of the subkeys, a warning is issued to notify
      the user about potential privilege escalation
    - Subkeys with invalid key binding signatures are discarded

  <Arguments>
    data:
          Public key data as written to stdout by
          securesystemslib.gpg.constants.GPG_EXPORT_PUBKEY_COMMAND.

    keyid:
          The keyid of the master key or one of its subkeys expected to be
          contained in the passed gpg data.

  <Exceptions>
    securesystemslib.gpg.exceptions.PacketParsingError
          If the key data could not be parsed

    securesystemslib.gpg.exceptions.KeyNotFoundError
          If the passed data is empty.
          If no master key or subkeys could be found that matches the passed
          keyid.

    securesystemslib.exceptions.FormatError
          If the passed keyid does not match
          securesystemslib.formats.KEYID_SCHEMA

  <Side Effects>
    None.

  <Returns>
    A public key in the format securesystemslib.formats.GPG_PUBKEY_SCHEMA with
    optional subkeys.

  """
  formats.KEYID_SCHEMA.check_match(keyid)
  if not data:
    raise KeyNotFoundError("Could not find gpg key '{}' in empty exported key "
        "data.".format(keyid))

  # Parse out master key and subkeys (enriched and verified via certificates
  # and binding signatures)
  raw_key_bundle = parse_pubkey_bundle(data)
  master_public_key = _assign_certified_key_info(raw_key_bundle)
  sub_public_keys = _get_verified_subkeys(raw_key_bundle)

  # Since GPG returns all pubkeys associated with a keyid (master key and
  # subkeys) we check which key matches the passed keyid.
  # If the matching key is a subkey, we warn the user because we return
  # the whole bundle (master plus all subkeys) and not only the subkey.
  # If no matching key is found we raise a KeyNotFoundError.
  for idx, public_key in enumerate(
      [master_public_key] + list(sub_public_keys.values())):
    if public_key and public_key["keyid"].endswith(keyid.lower()):
      if idx > 1:
        log.warning("Exporting master key '{}' including subkeys '{}' for"
            " passed keyid '{}'.".format(master_public_key["keyid"],
            ", ".join(list(sub_public_keys.keys())), keyid))
      break

  else:
    raise KeyNotFoundError("Could not find gpg key '{}' in exported key data."
        .format(keyid))

  # Add subkeys dictionary to master pubkey "subkeys" field if subkeys exist
  if sub_public_keys:
    master_public_key["subkeys"] = sub_public_keys

  return master_public_key


def parse_signature_packet(data, supported_signature_types=None,
    supported_hash_algorithms=None, include_info=False):
  """
  <Purpose>
    Parse the signature information on an RFC4880-encoded binary signature data
    buffer.

    NOTE: Older gpg versions (< FULLY_SUPPORTED_MIN_VERSION) might only
    reveal the partial key id. It is the callers responsibility to determine
    the full keyid based on the partial keyid, e.g. by exporting the related
    public and replacing the partial keyid with the full keyid.

  <Arguments>
    data:
           the RFC4880-encoded binary signature data buffer as described in
           section 5.2 (and 5.2.3.1).
    supported_signature_types: (optional)
          a set of supported signature_types, the signature packet may be
          (see securesystemslib.gpg.constants for available types). If None is
          specified the signature packet must be of type SIGNATURE_TYPE_BINARY.
    supported_hash_algorithms: (optional)
          a set of supported hash algorithm ids, the signature packet
          may use. Available ids are SHA1, SHA256, SHA512 (see
          securesystemslib.gpg.constants). If None is specified, the signature
          packet must use SHA256.
    include_info: (optional)
          a boolean that indicates whether an opaque dictionary should be
          added to the returned signature under the key "info". Default is
          False.

  <Exceptions>
    ValueError: if the signature packet is not supported or the data is
      malformed
    IndexError: if the signature packet is incomplete

  <Side Effects>
    None.

  <Returns>
    A signature dictionary matching
    securesystemslib.formats.GPG_SIGNATURE_SCHEMA with the following special
    characteristics:
     - The "keyid" field is an empty string if it cannot be determined
     - The "short_keyid" is not added if it cannot be determined
     - At least one of non-empty "keyid" or "short_keyid" are part of the
       signature

  """
  if not supported_signature_types:
    supported_signature_types = {SIGNATURE_TYPE_BINARY}

  if not supported_hash_algorithms:
    supported_hash_algorithms = {SHA256}

  _, header_len, _, packet_len = gpg_util.parse_packet_header(
      data, PACKET_TYPE_SIGNATURE)

  data = bytearray(data[header_len:packet_len])

  ptr = 0

  # we get the version number, which we also expect to be v4, or we bail
  # FIXME: support v3 type signatures (which I haven't seen in the wild)
  version_number = data[ptr]
  ptr += 1
  if version_number not in SUPPORTED_SIGNATURE_PACKET_VERSIONS:
    raise ValueError("Signature version '{}' not supported, must be one of "
        "{}.".format(version_number, SUPPORTED_SIGNATURE_PACKET_VERSIONS))

  # Per default we only parse "signatures of a binary document". Other types
  # may be allowed by passing type constants via `supported_signature_types`.
  # Types include revocation signatures, key binding signatures, persona
  # certifications, etc. (see RFC 4880 section 5.2.1.).
  signature_type = data[ptr]
  ptr += 1

  if signature_type not in supported_signature_types:
    raise ValueError("Signature type '{}' not supported, must be one of {} "
        "(see RFC4880 5.2.1. Signature Types).".format(signature_type,
        supported_signature_types))

  signature_algorithm = data[ptr]
  ptr += 1

  if signature_algorithm not in SUPPORTED_SIGNATURE_ALGORITHMS:
    raise ValueError("Signature algorithm '{}' not "
        "supported, please verify that your gpg configuration is creating "
        "either DSA, RSA, or EdDSA signatures (see RFC4880 9.1. Public-Key "
        "Algorithms).".format(signature_algorithm))

  key_type = SUPPORTED_SIGNATURE_ALGORITHMS[signature_algorithm]['type']
  handler = SIGNATURE_HANDLERS[key_type]

  hash_algorithm = data[ptr]
  ptr += 1

  if hash_algorithm not in supported_hash_algorithms:
    raise ValueError("Hash algorithm '{}' not supported, must be one of {}"
        " (see RFC4880 9.4. Hash Algorithms).".format(hash_algorithm,
        supported_hash_algorithms))

  # Obtain the hashed octets
  hashed_octet_count = struct.unpack(">H", data[ptr:ptr+2])[0]
  ptr += 2
  hashed_subpackets = data[ptr:ptr+hashed_octet_count]
  hashed_subpacket_info = gpg_util.parse_subpackets(hashed_subpackets)

  # Check whether we were actually able to read this much hashed octets
  if len(hashed_subpackets) != hashed_octet_count: # pragma: no cover
    raise ValueError("This signature packet seems to be corrupted."
        "It is missing hashed octets!")

  ptr += hashed_octet_count
  other_headers_ptr = ptr

  unhashed_octet_count = struct.unpack(">H", data[ptr: ptr + 2])[0]
  ptr += 2

  unhashed_subpackets = data[ptr:ptr+unhashed_octet_count]
  unhashed_subpacket_info = gpg_util.parse_subpackets(unhashed_subpackets)

  ptr += unhashed_octet_count

  # Use the info dict to return further signature information that may be
  # needed for intermediate processing, but does not have to be on the eventual
  # signature datastructure
  info = {
    "signature_type": signature_type,
    "hash_algorithm": hash_algorithm,
    "creation_time": None,
    "subpackets": {},
  }

  keyid = ""
  short_keyid = ""


  # Parse "Issuer" (short keyid) and "Issuer Fingerprint" (full keyid) type
  # subpackets
  # Strategy: Loop over all unhashed and hashed subpackets (in that order!) and
  # store only the last of a type. Due to the order in the loop, hashed
  # subpackets are prioritized over unhashed subpackets (see NOTEs below).

  # NOTE: A subpacket may be found either in the hashed or unhashed subpacket
  # sections of a signature. If a subpacket is not hashed, then the information
  # in it cannot be considered definitive because it is not part of the
  # signature proper. (see RFC4880 5.2.3.2.)
  # NOTE: Signatures may contain conflicting information in subpackets. In most
  # cases, an implementation SHOULD use the last subpacket, but MAY use any
  # conflict resolution scheme that makes more sense. (see RFC4880 5.2.4.1.)
  for idx, subpacket_tuple in \
      enumerate(unhashed_subpacket_info + hashed_subpacket_info):

    # The idx indicates if the info is from the unhashed (first) or
    # hashed (second) of the above concatenated lists
    is_hashed = (idx >= len(unhashed_subpacket_info))
    subpacket_type, subpacket_data = subpacket_tuple

    # Warn if expiration subpacket is not hashed
    if subpacket_type == KEY_EXPIRATION_SUBPACKET:
      if not is_hashed:
        log.warning("Expiration subpacket not hashed, gpg client possibly "
            "exporting a weakly configured key.")

    # Full keyids are only available in newer signatures
    # (see RFC4880 and rfc4880bis-06 5.2.3.1.)
    if subpacket_type == FULL_KEYID_SUBPACKET: # pragma: no cover
      # Exclude from coverage for consistent results across test envs
      # NOTE: The first byte of the subpacket payload is a version number
      # (see rfc4880bis-06 5.2.3.28.)
      keyid = binascii.hexlify(subpacket_data[1:]).decode("ascii")

    # We also return the short keyid, because the full might not be available
    if subpacket_type == PARTIAL_KEYID_SUBPACKET:
      short_keyid = binascii.hexlify(subpacket_data).decode("ascii")

    if subpacket_type == SIG_CREATION_SUBPACKET:
      info["creation_time"] = struct.unpack(">I", subpacket_data)[0]

    info["subpackets"][subpacket_type] = subpacket_data


  # Fail if there is no keyid at all (this should not happen)
  if not (keyid or short_keyid): # pragma: no cover
    raise ValueError("This signature packet seems to be corrupted. It does "
        "not have an 'Issuer' or 'Issuer Fingerprint' subpacket (see RFC4880 "
        "and rfc4880bis-06 5.2.3.1. Signature Subpacket Specification).")

  # Fail if keyid and short keyid are specified but don't match
  if keyid and not keyid.endswith(short_keyid): # pragma: no cover
    raise ValueError("This signature packet seems to be corrupted. The key ID "
        "'{}' of the 'Issuer' subpacket must match the lower 64 bits of the "
        "fingerprint '{}' of the 'Issuer Fingerprint' subpacket (see RFC4880 "
        "and rfc4880bis-06 5.2.3.28. Issuer Fingerprint).".format(
        short_keyid, keyid))

  if not info["creation_time"]: # pragma: no cover
    raise ValueError("This signature packet seems to be corrupted. It does "
        "not have a 'Signature Creation Time' subpacket (see RFC4880 5.2.3.4 "
        "Signature Creation Time).")

  # Uncomment this variable to obtain the left-hash-bits information (used for
  # early rejection)
  #left_hash_bits = struct.unpack(">H", data[ptr:ptr+2])[0]
  ptr += 2

  # Finally, fetch the actual signature (as opposed to signature metadata).
  signature = handler.get_signature_params(data[ptr:])

  signature_data = {
    'keyid': "{}".format(keyid),
    'other_headers': binascii.hexlify(data[:other_headers_ptr]).decode('ascii'),
    'signature': binascii.hexlify(signature).decode('ascii')
  }

  if short_keyid: # pragma: no branch
    signature_data["short_keyid"] = short_keyid

  if include_info:
    signature_data["info"] = info

  return signature_data



#!/usr/bin/env python

"""
<Program Name>
  check_public_interfaces.py

<Author>
  Joshua Lock <jlock@vmware.com>

<Started>
  January 6, 2020.

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Public facing modules (e.g. interface.py and keys.py) must be
  importable, even if the optional dependencies are not installed.

  Each public facing function should always be callable and present
  meaningful user-feedback if an optional dependency that is required for
  that function is not installed.

  This test purposefully only checks the public functions with a native
  dependency, to avoid duplicated tests.

  NOTE: the filename is purposefully check_ rather than test_ so that test
  discovery doesn't find this unittest and the tests within are only run
  when explicitly invoked.
"""

from __future__ import print_function
from __future__ import absolute_import

import inspect
import json
import os
import shutil
import sys
import tempfile
import unittest

if sys.version_info >= (3, 3):
  import unittest.mock as mock
else:
  import mock

import securesystemslib.exceptions
import securesystemslib.gpg.constants
import securesystemslib.gpg.functions
import securesystemslib.gpg.util
import securesystemslib.interface
import securesystemslib.keys



class TestPublicInterfaces(unittest.TestCase):

  @classmethod
  def setUpClass(cls):
    cls.temp_dir = tempfile.mkdtemp(dir=os.getcwd())

  @classmethod
  def tearDownClass(cls):
    shutil.rmtree(cls.temp_dir)

  def test_interface(self):

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.interface._generate_and_write_rsa_keypair(password='pw')

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.interface.generate_and_write_rsa_keypair('pw')

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.interface.generate_and_write_rsa_keypair('pw')

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      # Mock entry on prompt which is presented before lower-level functions
      # raise UnsupportedLibraryError
      with mock.patch("securesystemslib.interface.get_password", return_value=""):
        securesystemslib.interface.generate_and_write_rsa_keypair_with_prompt()

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.interface.generate_and_write_unencrypted_rsa_keypair()

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      path = os.path.join(self.temp_dir, 'rsa_key')
      with open(path, 'a'):
        securesystemslib.interface.import_rsa_privatekey_from_file(
            path)

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.interface._generate_and_write_ed25519_keypair(
          password='pw')

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.interface.generate_and_write_ed25519_keypair('pw')

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      # Mock entry on prompt which is presented before lower-level functions
      # raise UnsupportedLibraryError
      with mock.patch("securesystemslib.interface.get_password", return_value=""):
        securesystemslib.interface.generate_and_write_ed25519_keypair_with_prompt()

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.interface.generate_and_write_unencrypted_ed25519_keypair()

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      path = os.path.join(self.temp_dir, 'ed25519_priv.json')
      with open(path, 'a') as f:
        f.write('{}')
        securesystemslib.interface.import_ed25519_privatekey_from_file(
            path, 'pw')

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.interface._generate_and_write_ecdsa_keypair(
          password='pw')

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.interface.generate_and_write_ecdsa_keypair('pw')

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      # Mock entry on prompt which is presented before lower-level functions
      # raise UnsupportedLibraryError
      with mock.patch("securesystemslib.interface.get_password", return_value=""):
        securesystemslib.interface.generate_and_write_ecdsa_keypair_with_prompt()

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.interface.generate_and_write_unencrypted_ecdsa_keypair()

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      path = os.path.join(self.temp_dir, 'ecddsa.priv')
      with open(path, 'a') as f:
        f.write('{}')
        securesystemslib.interface.import_ecdsa_privatekey_from_file(
            path, password='pw')


  def test_keys(self):
    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.keys.generate_rsa_key()

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.keys.generate_ecdsa_key()

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.keys.generate_ed25519_key()

    data = 'foo'
    keydict = {'keytype': 'ed25519',
               'scheme': 'ed25519',
               'keyid': 'f00',
               'keyval': {'private': 'f001',
                          'public': 'b00f'}}
    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.keys.create_signature(keydict, data)

    keydict['keytype'] = 'ecdsa'
    keydict['scheme'] = 'ecdsa-sha2-nistp256'
    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.keys.create_signature(keydict, data)

    keydict['keytype'] = 'rsa'
    keydict['scheme'] = 'rsassa-pss-sha256'
    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.keys.create_signature(keydict, data)

    keydict['keytype'] = 'ecdsa'
    keydict['scheme'] = 'ecdsa-sha2-nistp256'
    sig = {'keyid': 'f00',
           'sig': 'cfbce8e23eef478975a4339036de2335002d57c7b1632dd01e526a3bc52a5b261508ad50b9e25f1b819d61017e7347e912db1af019bf47ee298cc58bbdef9703'}
    # NOTE: we don't test ed25519 keys as they can be verified in pure python
    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.keys.verify_signature(keydict, sig, data)

    keydict['keytype'] = 'rsa'
    keydict['scheme'] = 'rsassa-pss-sha256'
    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.keys.verify_signature(keydict, sig, data)

    priv = '-----BEGIN RSA PRIVATE KEY-----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\nMZW81M7ahgizqGQsRcOskA/uBC1w3N7o/lUYa3I+OY6EqA4KigIuGw==\n-----END RSA PRIVATE KEY-----\n'

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.keys.import_rsakey_from_private_pem('')

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.keys.encrypt_key(keydict, 'foo')

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.keys.decrypt_key('enc', 'pw')

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.keys.create_rsa_encrypted_pem(priv, 'pw')

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.keys.import_ed25519key_from_private_json(
          ''.encode('utf-8'), '')

    with self.assertRaises(
          securesystemslib.exceptions.UnsupportedLibraryError):
      securesystemslib.keys.import_ecdsakey_from_private_pem(priv)

  def test_purepy_ed25519(self):
    data = b'The quick brown fox jumps over the lazy dog'
    pub = b'\xbe\xb7\\&\x82\x06UN\x96<E\xdc\xbf<\x00A@\xd1\xcbi\xbb\xfe\x93p\xefso\x198\x8c\x9b&'
    sig = b'.\xc7\xa5\xe2\x95\xfabe\xe1\x0f=\xa7\xf1\xa42\xe7t/\x04\x1f\x08\x1bO\xae\xca\xb3\xa1+\xf0\xfc\x8f6l\x91\x9c\x90\xc2g\xe9\xed\x1d\xfd\xebzuV\xb9Y\xa9m\xd0\xdc\xfe\xa1}\xa3Xb-9\xaf6\xbf\t'

    valid = securesystemslib.ed25519_keys.verify_signature(
        pub, 'ed25519', sig, data)
    self.assertEqual(True, valid)

    bsig = b'\xd3/\x7f\x7f\xa5;6Pq\x14f]\x8b\x0e@\x8a:\xc2\xa1\xb8\xee\x11\xef\x06s\x12\xa9\x0b0\xe9@\xd5Q\xb6\xf7\xe7\xb9\xf6\xc7J\x99_L\x01\xf7\xcdi\x05\xea\xdf\x05D\x12\x1f\xeeT\xe1y\xb1\x9a\x8e\xebS\x04'
    invalid = securesystemslib.ed25519_keys.verify_signature(
        pub, 'ed25519', bsig, data)
    self.assertEqual(False, invalid)

  def test_gpg_functions(self):
    """Public GPG functions must raise error on missing cryptography lib. """
    expected_error = securesystemslib.exceptions.UnsupportedLibraryError
    expected_error_msg = securesystemslib.gpg.functions.NO_CRYPTO_MSG

    with self.assertRaises(expected_error) as ctx:
      securesystemslib.gpg.functions.create_signature('bar')
    self.assertEqual(expected_error_msg, str(ctx.exception))

    with self.assertRaises(expected_error) as ctx:
      securesystemslib.gpg.functions.verify_signature(None, 'f00', 'bar')
    self.assertEqual(expected_error_msg, str(ctx.exception))

    with self.assertRaises(expected_error) as ctx:
      securesystemslib.gpg.functions.export_pubkey('f00')
    self.assertEqual(expected_error_msg, str(ctx.exception))

if __name__ == "__main__":
  unittest.main(verbosity=1, buffer=True)



#!/usr/bin/env python

"""
<Program Name>
  schema.py

<Author>
  Geremy Condra
  Vladimir Diaz <vladimir.v.diaz@gmail.com>

<Started>
  Refactored April 30, 2012 (previously named checkjson.py). -Vlad

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provide a variety of classes that compare objects
  based on their format and determine if they match.
  These classes, or schemas, do not simply check the
  type of the objects being compared, but inspect
  additional aspects of the objects like names and
  the number of items included.
  For example:
  >>> good = {'first': 'Marty', 'last': 'McFly'}
  >>> bad = {'sdfsfd': 'Biff', 'last': 'Tannen'}
  >>> bad = {'sdfsfd': 'Biff', 'last': 'Tannen'}
  >>> schema = Object(first=AnyString(), last=AnyString())
  >>> schema.matches(good)
  True
  >>> schema.matches(bad)
  False

  In the process of determining if the two objects matched the template,
  securesystemslib.schema.Object() inspected the named keys of both
  dictionaries.  In the case of the 'bad' dict, a 'first' dict key could not be
  found.  As a result, 'bad' was flagged a mismatch.

  'schema.py' provides additional schemas for testing objects based on other
  criteria.  See 'securesystemslib.formats.py' and the rest of this module for
  extensive examples.  Anything related to the checking of securesystemslib
  objects and their formats can be found in 'formats.py'.
"""

# Help with Python 3 compatibility, where the print statement is a function, an
# implicit relative import is invalid, and the '/' operator performs true
# division.  Example:  print 'hello world' raises a 'SyntaxError' exception.
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import re
import sys

import six

from securesystemslib import exceptions


class Schema:
  """
  <Purpose>
    A schema matches a set of possible Python objects, of types
    that are encodable in JSON.  'Schema' is the base class for
    the other classes defined in this module.  All derived classes
    should implement check_match().
  """

  def matches(self, object):
    """
    <Purpose>
      Return True if 'object' matches this schema, False if it doesn't.
      If the caller wishes to signal an error on a failed match, check_match()
      should be called, which will raise a 'exceptions.FormatError' exception.
    """

    try:
      self.check_match(object)
    except exceptions.FormatError:
      return False
    else:
      return True


  def check_match(self, object):
    """
    <Purpose>
      Abstract method.  Classes that inherit from 'Schema' must
      implement check_match().  If 'object' matches the schema, check_match()
      should simply return.  If 'object' does not match the schema,
      'exceptions.FormatError' should be raised.
    """

    raise NotImplementedError()





class Any(Schema):
  """
  <Purpose>
    Matches any single object.  Whereas other schemas explicitly state
    the required type of its argument, Any() does not. It simply does a
    'pass' when 'check_match()' is called and at the point where the schema
    is instantiated.

    Supported methods include:
      matches(): returns a Boolean result.
      check_match(): passed

  <Example Use>

    >>> schema = Any()
    >>> schema.matches('A String')
    True
    >>> schema.matches([1, 'list'])
    True
  """

  def __init__(self):
    pass


  def check_match(self, object):
    pass





class String(Schema):
  """
  <Purpose>
    Matches a particular string.  The argument object must be a string and be
    equal to a specific string value.  At instantiation, the string is set and
    any future comparisons are checked against this internal string value.

    Supported methods include:
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.

  <Example Use>

    >>> schema = String('Hi')
    >>> schema.matches('Hi')
    True
    >>> schema.matches('Not hi')
    False
  """

  def __init__(self, string):
    if not isinstance(string, six.string_types):
      raise exceptions.FormatError('Expected a string but'
          ' got ' + repr(string))

    self._string = string


  def check_match(self, object):
    if self._string != object:
      raise exceptions.FormatError(
          'Expected ' + repr(self._string) + ' got ' + repr(object))





class AnyString(Schema):
  """
  <Purpose>
    Matches any string, but not a non-string object.  This schema
    can be viewed as the Any() schema applied to Strings, but an
    additional check is performed to ensure only strings are considered.

    Supported methods include:
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.

  <Example Use>

    >>> schema = AnyString()
    >>> schema.matches('')
    True
    >>> schema.matches('a string')
    True
    >>> schema.matches(['a'])
    False
    >>> schema.matches(3)
    False
    >>> schema.matches(u'a unicode string')
    True
    >>> schema.matches({})
    False
  """

  def __init__(self):
    pass


  def check_match(self, object):
    if not isinstance(object, six.string_types):
      raise exceptions.FormatError('Expected a string'
          ' but got ' + repr(object))





class AnyNonemptyString(AnyString):
  """
  <Purpose>
    Matches any string with one or more characters.
    This schema can be viewed as the Any() schema applied to Strings, but an
    additional check is performed to ensure only strings are considered and
    that said strings have at least one character.

    Supported methods include:
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.

  <Example Use>

    >>> schema = AnyNonemptyString()
    >>> schema.matches('')
    False
    >>> schema.matches('a string')
    True
    >>> schema.matches(['a'])
    False
    >>> schema.matches(3)
    False
    >>> schema.matches(u'a unicode string')
    True
    >>> schema.matches({})
    False
  """

  def check_match(self, object):
    AnyString.check_match(self, object)

    if object == "":
        raise exceptions.FormatError('Expected a string'
            ' with at least one character but got ' + repr(object))





class AnyBytes(Schema):
  """
  <Purpose>
    Matches any byte string, but not a non-byte object.  This schema can be
    viewed as the Any() schema applied to byte strings, but an additional check
    is performed to ensure only strings are considered.  Supported methods
    include:
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.

  <Example Use>

    >>> schema = AnyBytes()
    >>> schema.matches(b'')
    True
    >>> schema.matches(b'a string')
    True
    >>> schema.matches(['a'])
    False
    >>> schema.matches(3)
    False
    >>> schema.matches({})
    False
  """

  def __init__(self):
    pass


  def check_match(self, object):
    if not isinstance(object, six.binary_type):
      raise exceptions.FormatError('Expected a byte string'
          ' but got ' + repr(object))





class LengthString(Schema):
  """
  <Purpose>
    Matches any string of a specified length.  The argument object must be a
    string.  At instantiation, the string length is set and any future
    comparisons are checked against this internal string value length.

    Supported methods include:
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.

  <Example Use>

    >>> schema = LengthString(5)
    >>> schema.matches('Hello')
    True
    >>> schema.matches('Hi')
    False
  """

  def __init__(self, length):
    if isinstance(length, bool) or not isinstance(length, six.integer_types):
      # We need to check for bool as a special case, since bool
      # is for historical reasons a subtype of int.
      raise exceptions.FormatError(
          'Got ' + repr(length) + ' instead of an integer.')

    self._string_length = length


  def check_match(self, object):
    if not isinstance(object, six.string_types):
      raise exceptions.FormatError('Expected a string but'
          ' got ' + repr(object))

    if len(object) != self._string_length:
      raise exceptions.FormatError('Expected a string of'
          ' length ' + repr(self._string_length))





class LengthBytes(Schema):
  """
  <Purpose>
    Matches any Bytes of a specified length.  The argument object must be either
    a str() in Python 2, or bytes() in Python 3.  At instantiation, the bytes
    length is set and any future comparisons are checked against this internal
    bytes value length.

    Supported methods include:
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.

  <Example Use>

    >>> schema = LengthBytes(5)
    >>> schema.matches(b'Hello')
    True
    >>> schema.matches(b'Hi')
    False
  """

  def __init__(self, length):
    if isinstance(length, bool) or not isinstance(length, six.integer_types):
      # We need to check for bool as a special case, since bool
      # is for historical reasons a subtype of int.
      raise exceptions.FormatError(
          'Got ' + repr(length) + ' instead of an integer.')

    self._bytes_length = length


  def check_match(self, object):
    if not isinstance(object, six.binary_type):
      raise exceptions.FormatError('Expected a byte but'
          ' got ' + repr(object))

    if len(object) != self._bytes_length:
      raise exceptions.FormatError('Expected a byte of'
          ' length ' + repr(self._bytes_length))





class OneOf(Schema):
  """
  <Purpose>
    Matches an object that matches any one of several schemas.  OneOf() returns
    a result as soon as one of its recognized sub-schemas is encountered in the
    object argument.  When OneOf() is instantiated, its supported sub-schemas
    are specified by a sequence type (e.g., a list, tuple, etc.).  A mismatch
    is returned after checking all sub-schemas and not finding a supported
    type.

    Supported methods include:
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.

  <Example Use>
    >>> schema = OneOf([ListOf(Integer()), String('Hello'), String('bye')])
    >>> schema.matches(3)
    False
    >>> schema.matches('bye')
    True
    >>> schema.matches([])
    True
    >>> schema.matches([1,2])
    True
    >>> schema.matches(['Hi'])
    False
  """

  def __init__(self, alternatives):
    # Ensure each item of the list contains the expected object type.
    if not isinstance(alternatives, list):
      raise exceptions.FormatError('Expected a list but'
          ' got ' + repr(alternatives))

    for alternative in alternatives:
      if not isinstance(alternative, Schema):
        raise exceptions.FormatError('List contains an'
            ' invalid item ' + repr(alternative))

    self._alternatives = alternatives


  def check_match(self, object):
    # Simply return as soon as we find a match.
    # Raise 'exceptions.FormatError' if no matches are found.
    for alternative in self._alternatives:
      if alternative.matches(object):
        return
    raise exceptions.FormatError('Object did not match a'
        ' recognized alternative.')





class AllOf(Schema):
  """
  <Purpose>
    Matches the intersection of a list of schemas.  The object being tested
    must match all of the required sub-schemas.  Unlike OneOf(), which can
    return a result as soon as a match is found in one of its supported
    sub-schemas, AllOf() must verify each sub-schema before returning a result.

    Supported methods include:
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.

  <Example Use>
    >>> schema = AllOf([Any(), AnyString(), String('a')])
    >>> schema.matches('b')
    False
    >>> schema.matches('a')
    True
  """

  def __init__(self, required_schemas):
    # Ensure each item of the list contains the expected object type.
    if not isinstance(required_schemas, list):
      raise exceptions.FormatError('Expected a list but'
          ' got' + repr(required_schemas))

    for schema in required_schemas:
      if not isinstance(schema, Schema):
        raise exceptions.FormatError('List contains an'
            ' invalid item ' + repr(schema))

    self._required_schemas = required_schemas[:]


  def check_match(self, object):
    for required_schema in self._required_schemas:
      required_schema.check_match(object)





class Boolean(Schema):
  """
  <Purpose>
    Matches a boolean.  The object argument must be one of True or False.  All
    other types are flagged as mismatches.

    Supported methods include:
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.

  <Example Use>
    >>> schema = Boolean()
    >>> schema.matches(True) and schema.matches(False)
    True
    >>> schema.matches(11)
    False
 """

  def __init__(self):
    pass


  def check_match(self, object):
    if not isinstance(object, bool):
      raise exceptions.FormatError(
          'Got ' + repr(object) + ' instead of a boolean.')





class ListOf(Schema):
  """
  <Purpose>
    Matches a homogeneous list of some sub-schema.  That is, all the sub-schema
    must be of the same type.  The object argument must be a sequence type
    (e.g., a list, tuple, etc.).  When ListOf() is instantiated, a minimum and
    maximum count can be specified for the homogeneous sub-schema list.  If
    min_count is set to 'n', the object argument sequence must contain 'n'
    items.  See ListOf()'s __init__ method for the expected arguments.

    Supported methods include:
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.

  <Example Use>
    >>> schema = ListOf(RegularExpression('(?:..)*'))
    >>> schema.matches('hi')
    False
    >>> schema.matches([])
    True
    >>> schema.matches({})
    False
    >>> schema.matches(['Hi', 'this', 'list', 'is', 'full', 'of', 'even', 'strs'])
    True
    >>> schema.matches(['This', 'one', 'is not'])
    False
    >>> schema = ListOf(Integer(), min_count=3, max_count=10)
    >>> schema.matches([3]*2)
    False
    >>> schema.matches([3]*3)
    True
    >>> schema.matches([3]*10)
    True
    >>> schema.matches([3]*11)
    False
  """

  def __init__(self, schema, min_count=0, max_count=sys.maxsize, list_name='list'):
    """
    <Purpose>
      Create a new ListOf schema.

    <Arguments>
      schema:  The pattern to match.
      min_count: The minimum number of sub-schema in 'schema'.
      max_count: The maximum number of sub-schema in 'schema'.
      list_name: A string identifier for the ListOf object.
    """

    if not isinstance(schema, Schema):
      message = 'Expected Schema type but got '+repr(schema)
      raise exceptions.FormatError(message)

    self._schema = schema
    self._min_count = min_count
    self._max_count = max_count
    self._list_name = list_name


  def check_match(self, object):
    if not isinstance(object, (list, tuple)):
      raise exceptions.FormatError(
          'Expected object of type {} but got type {}'.format(
            self._list_name, type(object).__name__))


    # Check if all the items in the 'object' list
    # match 'schema'.
    for item in object:
      try:
        self._schema.check_match(item)

      except exceptions.FormatError as e:
        raise exceptions.FormatError(
            str(e) + ' in ' + repr(self._list_name))

    # Raise exception if the number of items in the list is
    # not within the expected range.
    if not (self._min_count <= len(object) <= self._max_count):
        raise exceptions.FormatError(
            'Length of ' + repr(self._list_name) + ' out of range.')





class Integer(Schema):
  """
  <Purpose>
    Matches an integer.  A range can be specified.  For example, only integers
    between 8 and 42 can be set as a requirement.  The object argument is also
    checked against a Boolean type, since booleans have historically been
    considered a sub-type of integer.

    Supported methods include:
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.

  <Example Use>
    >>> schema = Integer()
    >>> schema.matches(99)
    True
    >>> schema.matches(False)
    False
    >>> schema.matches('a string')
    False
    >>> Integer(lo=10, hi=30).matches(25)
    True
    >>> Integer(lo=10, hi=30).matches(5)
    False
  """

  def __init__(self, lo = -2147483648, hi = 2147483647):
    """
    <Purpose>
      Create a new Integer schema.

    <Arguments>
      lo: The minimum value the int object argument can be.
      hi: The maximum value the int object argument can be.
    """

    self._lo = lo
    self._hi = hi


  def check_match(self, object):
    if isinstance(object, bool) or not isinstance(object, six.integer_types):
      # We need to check for bool as a special case, since bool
      # is for historical reasons a subtype of int.
      raise exceptions.FormatError(
          'Got ' + repr(object) + ' instead of an integer.')

    elif not (self._lo <= object <= self._hi):
      int_range = '[' + repr(self._lo) + ', ' + repr(self._hi) + '].'
      raise exceptions.FormatError(
          repr(object) + ' not in range ' + int_range)





class DictOf(Schema):
  """
  <Purpose>
    Matches a mapping from items matching a particular key-schema to items
    matching a value-schema (i.e., the object being checked must be a dict).
    Note that in JSON, keys must be strings.  In the example below, the keys of
    the dict must be one of the letters contained in 'aeiou' and the value must
    be a structure containing any two strings.

    Supported methods include:
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.

  <Example Use>
    >>> schema = DictOf(RegularExpression(r'[aeiou]+'), Struct([AnyString(), AnyString()]))
    >>> schema.matches('')
    False
    >>> schema.matches({})
    True
    >>> schema.matches({'a': ['x', 'y'], 'e' : ['', '']})
    True
    >>> schema.matches({'a': ['x', 3], 'e' : ['', '']})
    False
    >>> schema.matches({'a': ['x', 'y'], 'e' : ['', ''], 'd' : ['a', 'b']})
    False
  """

  def __init__(self, key_schema, value_schema):
    """
    <Purpose>
      Create a new DictOf schema.

    <Arguments>
      key_schema:  The dictionary's key.
      value_schema: The dictionary's value.
    """

    if not isinstance(key_schema, Schema):
      raise exceptions.FormatError('Expected Schema but'
          ' got ' + repr(key_schema))

    if not isinstance(value_schema, Schema):
      raise exceptions.FormatError('Expected Schema but'
          ' got ' + repr(value_schema))

    self._key_schema = key_schema
    self._value_schema = value_schema


  def check_match(self, object):
    if not isinstance(object, dict):
      raise exceptions.FormatError('Expected a dict but'
          ' got ' + repr(object))

    for key, value in six.iteritems(object):
      self._key_schema.check_match(key)
      self._value_schema.check_match(value)





class Optional(Schema):
  """
  <Purpose>
    Provide a way for the Object() schema to accept optional dictionary keys.
    The Object() schema outlines how a dictionary should look, such as the
    names for dict keys and the object type of the dict values.  Optional()'s
    intended use is as a sub-schema to Object().  Object() flags an object as a
    mismatch if a required key is not encountered, however, dictionary keys
    labeled Optional() are not required to appear in the object's list of
    required keys.  If an Optional() key IS found, Optional()'s sub-schemas are
    then verified.

    Supported methods include:
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.

  <Example Use>
    >>> schema = Object(k1=String('X'), k2=Optional(String('Y')))
    >>> schema.matches({'k1': 'X', 'k2': 'Y'})
    True
    >>> schema.matches({'k1': 'X', 'k2': 'Z'})
    False
    >>> schema.matches({'k1': 'X'})
    True
  """

  def __init__(self, schema):
    if not isinstance(schema, Schema):
      raise exceptions.FormatError('Expected Schema, but'
          ' got ' + repr(schema))
    self._schema = schema


  def check_match(self, object):
    self._schema.check_match(object)





class Object(Schema):
  """
  <Purpose>
    Matches a dict from specified keys to key-specific types.  Unrecognized
    keys are allowed.  The Object() schema outlines how a dictionary should
    look, such as the names for dict keys and the object type of the dict
    values.  See schema.Optional() to learn how Object() incorporates optional
    sub-schemas.

    Supported methods include:
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.

  <Example Use>
    >>> schema = Object(a=AnyString(), bc=Struct([Integer(), Integer()]))
    >>> schema.matches({'a':'ZYYY', 'bc':[5,9]})
    True
    >>> schema.matches({'a':'ZYYY', 'bc':[5,9], 'xx':5})
    True
    >>> schema.matches({'a':'ZYYY', 'bc':[5,9,3]})
    False
    >>> schema.matches({'a':'ZYYY'})
    False
  """

  def __init__(self, object_name='object', **required):
    """
    <Purpose>
      Create a new Object schema.

    <Arguments>
      object_name: A string identifier for the object argument.

      A variable number of keyword arguments is accepted.
    """

    # Ensure valid arguments.
    for key, schema in six.iteritems(required):
      if not isinstance(schema, Schema):
        raise exceptions.FormatError('Expected Schema but'
            ' got ' + repr(schema))

    self._object_name = object_name
    self._required = list(required.items())


  def check_match(self, object):
    if not isinstance(object, dict):
      raise exceptions.FormatError(
          'Wanted a ' + repr(self._object_name) + '.')

    # (key, schema) = (a, AnyString()) = (a=AnyString())
    for key, schema in self._required:
      # Check if 'object' has all the required dict keys.  If not one of the
      # required keys, check if it is an Optional().
      try:
        item = object[key]

      except KeyError:
        # If not an Optional schema, raise an exception.
        if not isinstance(schema, Optional):
          raise exceptions.FormatError(
              'Missing key ' + repr(key) + ' in ' + repr(self._object_name))

      # Check that 'object's schema matches Object()'s schema for this
      # particular 'key'.
      else:
        try:
          schema.check_match(item)

        except exceptions.FormatError as e:
          raise exceptions.FormatError(
              str(e) + ' in ' + self._object_name + '.' + key)





class Struct(Schema):
  """
  <Purpose>
    Matches a non-homogeneous list of items.  The sub-schemas are allowed to
    vary.  The object argument must be a sequence type (e.g., a list, tuple,
    etc.).  There is also an option to specify that additional schemas not
    explicitly defined at instantiation are allowed.  See __init__() for the
    complete list of arguments accepted.

    Supported methods include:
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.

  <Example Use>
    >>> schema = Struct([ListOf(AnyString()), AnyString(), String('X')])
    >>> schema.matches(False)
    False
    >>> schema.matches('Foo')
    False
    >>> schema.matches([[], 'Q', 'X'])
    True
    >>> schema.matches([[], 'Q', 'D'])
    False
    >>> schema.matches([[3], 'Q', 'X'])
    False
    >>> schema.matches([[], 'Q', 'X', 'Y'])
    False
    >>> schema = Struct([String('X')], allow_more=True)
    >>> schema.matches([])
    False
    >>> schema.matches(['X'])
    True
    >>> schema.matches(['X', 'Y'])
    True
    >>> schema.matches(['X', ['Y', 'Z']])
    True
    >>> schema.matches([['X']])
    False
    >>> schema = Struct([String('X'), Integer()], [Integer()])
    >>> schema.matches([])
    False
    >>> schema.matches({})
    False
    >>> schema.matches(['X'])
    False
    >>> schema.matches(['X', 3])
    True
    >>> schema.matches(['X', 3, 9])
    True
    >>> schema.matches(['X', 3, 9, 11])
    False
    >>> schema.matches(['X', 3, 'A'])
    False
  """

  def __init__(self, sub_schemas, optional_schemas=None, allow_more=False,
               struct_name='list'):
    """
    <Purpose>
      Create a new Struct schema.

    <Arguments>
      sub_schemas: The sub-schemas recognized.
      optional_schemas: Optional list. If none is given, it will be "[]".
      allow_more: Specifies that an optional list of types is allowed.
      struct_name: A string identifier for the Struct object.
    """

    if optional_schemas is None:
      optional_schemas = []

    # Ensure each item of the list contains the expected object type.
    if not isinstance(sub_schemas, (list, tuple)):
      raise exceptions.FormatError(
          'Expected Schema but got ' + repr(sub_schemas))

    for schema in sub_schemas:
      if not isinstance(schema, Schema):
        raise exceptions.FormatError('Expected Schema but'
            ' got ' + repr(schema))

    self._sub_schemas = sub_schemas + optional_schemas
    self._min = len(sub_schemas)
    self._allow_more = allow_more
    self._struct_name = struct_name


  def check_match(self, object):
    if not isinstance(object, (list, tuple)):
      raise exceptions.FormatError(
          'Expected ' + repr(self._struct_name) + '; but got ' + repr(object))

    elif len(object) < self._min:
      raise exceptions.FormatError(
          'Too few fields in ' + self._struct_name)

    elif len(object) > len(self._sub_schemas) and not self._allow_more:
      raise exceptions.FormatError(
          'Too many fields in ' + self._struct_name)

    # Iterate through the items of 'object', checking against each schema in
    # the list of schemas allowed (i.e., the sub-schemas and also any optional
    # schemas.  The lenth of 'object' must be less than the length of the
    # required schemas + the optional schemas.  However, 'object' is allowed to
    # be only as large as the length of the required schemas.  In the while
    # loop below, we check against these two cases.
    index = 0
    while index < len(object) and index < len(self._sub_schemas):
      item = object[index]
      schema = self._sub_schemas[index]
      schema.check_match(item)
      index = index + 1





class RegularExpression(Schema):
  """
  <Purpose>
    Matches any string that matches a given regular expression.  The RE pattern
    set when RegularExpression is instantiated must not be None.  See
    __init__() for a complete list of accepted arguments.

    Supported methods include:
      matches(): returns a Boolean result.
      check_match(): raises 'exceptions.FormatError' on a mismatch.

  <Example Use>
    >>> schema = RegularExpression('h.*d')
    >>> schema.matches('hello world')
    True
    >>> schema.matches('Hello World')
    False
    >>> schema.matches('hello world!')
    False
    >>> schema.matches([33, 'Hello'])
    False
  """

  def __init__(self, pattern=None, modifiers=0, re_object=None, re_name=None):
    """
    <Purpose>
      Create a new regular expression schema.

    <Arguments>
      pattern:  The pattern to match, or None if re_object is provided.
      modifiers:  Flags to use when compiling the pattern.
      re_object:  A compiled regular expression object.
      re_name: Identifier for the regular expression object.
    """

    if not isinstance(pattern, six.string_types):
      if pattern is not None:
        raise exceptions.FormatError(
            repr(pattern) + ' is not a string.')

    if re_object is None:
      if pattern is None:
        raise exceptions.FormatError(
            'Cannot compare against an unset regular expression')

      if not pattern.endswith('$'):
        pattern += '$'
      re_object = re.compile(pattern, modifiers)
    self._re_object = re_object

    if re_name is None:
      if pattern is not None:
        re_name = 'pattern /' + pattern + '/'

      else:
        re_name = 'pattern'
    self._re_name = re_name


  def check_match(self, object):
    if not isinstance(object, six.string_types) or not self._re_object.match(object):
      raise exceptions.FormatError(
          repr(object) + ' did not match ' + repr(self._re_name))



if __name__ == '__main__':
  # The interactive sessions of the documentation strings can
  # be tested by running schema.py as a standalone module.
  # python -B schema.py.
  import doctest
  doctest.testmod()




"""
<Module Name>
  constants.py

<Author>
  Santiago Torres-Arias <santiago@nyu.edu>

<Started>
  Nov 15, 2017

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  aggregates all the constant definitions and lookup structures for signature
  handling
"""
import logging
import os

from securesystemslib import process

log = logging.getLogger(__name__)


def is_available_gnupg(gnupg):
  gpg_version_cmd = gnupg + " --version"
  try:
    process.run(gpg_version_cmd, stdout=process.PIPE, stderr=process.PIPE)
    return True
  except OSError:
    return False


GPG_COMMAND = ""
HAVE_GPG = False

GPG_ENV_COMMAND = os.environ.get('GNUPG')
GPG2_COMMAND = "gpg2"
GPG1_COMMAND = "gpg"

# By default, we allow providing GPG client through the environment
# assuming gpg2 as default value and test if exists. Otherwise, we assume gpg
# exists.
if GPG_ENV_COMMAND:
  if is_available_gnupg(GPG_ENV_COMMAND):
    GPG_COMMAND = GPG_ENV_COMMAND
elif is_available_gnupg(GPG2_COMMAND):
  GPG_COMMAND = GPG2_COMMAND
elif is_available_gnupg(GPG1_COMMAND):
  GPG_COMMAND = GPG1_COMMAND

if GPG_COMMAND:
  # Use bool to skip tests or fail early and gracefully if no gpg is available
  HAVE_GPG = True

GPG_VERSION_COMMAND = GPG_COMMAND + " --version"
FULLY_SUPPORTED_MIN_VERSION = "2.1.0"
NO_GPG_MSG = "GPG support requires a GPG client. 'gpg2' or 'gpg' with version {} or newer is" \
  " fully supported.".format(FULLY_SUPPORTED_MIN_VERSION)

GPG_SIGN_COMMAND = GPG_COMMAND + \
                   " --detach-sign --digest-algo SHA256 {keyarg} {homearg}"
GPG_EXPORT_PUBKEY_COMMAND = GPG_COMMAND + " {homearg} --export {keyid}"

# See RFC4880 section 4.3. Packet Tags for a list of all packet types The
# relevant packets defined below are described in sections 5.2 (signature),
# 5.5.1.1 (primary pubkey) and 5.5.1.2 (pub subkey), 5.12 (user id) and 5.13
# (user attribute)
PACKET_TYPE_SIGNATURE = 0x02
PACKET_TYPE_PRIMARY_KEY = 0x06
PACKET_TYPE_USER_ID = 0x0D
PACKET_TYPE_USER_ATTR = 0x11
PACKET_TYPE_SUB_KEY = 0x0E


# See sections 5.2.3 (signature) and 5.5.2 (public key) of RFC4880
SUPPORTED_SIGNATURE_PACKET_VERSIONS = {0x04}
SUPPORTED_PUBKEY_PACKET_VERSIONS = {0x04}

# The constants for hash algorithms are taken from section 9.4 of RFC4880.
SHA1 = 0x02
SHA256 = 0x08
SHA512 = 0x0A

# See section 5.2.1 of RFC4880
SIGNATURE_TYPE_BINARY = 0x00
SIGNATURE_TYPE_SUB_KEY_BINDING = 0x18
SIGNATURE_TYPE_CERTIFICATES = {0x10, 0x11, 0x12, 0x13}

# See section 5.2.3.4 (Signature Creation Time) of RFC4880
SIG_CREATION_SUBPACKET = 0x02
# See section 5.2.3.5. (Issuer) of RFC4880
PARTIAL_KEYID_SUBPACKET = 0x10
# See section 5.2.3.6 (Key Expiration Time) of RFC4880
KEY_EXPIRATION_SUBPACKET = 0x09
# See section 5.2.3.19 (Primary User ID) of RFC4880
PRIMARY_USERID_SUBPACKET = 0x19
# See section 5.2.3.28. (Issuer Fingerprint) of rfc4880bis-06
FULL_KEYID_SUBPACKET = 0x21



#!/usr/bin/env python

"""
<Program Name>
  formats.py

<Author>
  Geremy Condra
  Vladimir Diaz <vladimir.v.diaz@gmail.com>

<Started>
  Refactored April 30, 2012. -vladimir.v.diaz

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  A central location for all format-related checking of securesystemslib
  objects. Note: 'formats.py' depends heavily on 'schema.py', so the
  'schema.py' module should be read and understood before tackling this module.

  'formats.py' can be broken down into three sections.  (1) Schemas and object
  matching.  (2) Classes that represent Role Metadata and help produce
  correctly formatted files.  (3) Functions that help produce or verify
  securesystemslib objects.

  The first section deals with schemas and object matching based on format.
  There are two ways of checking the format of objects.  The first method
  raises a 'securesystemslib.exceptions.FormatError' exception if the match
  fails and the other returns a Boolean result.

  securesystemslib.formats.<SCHEMA>.check_match(object)
  securesystemslib.formats.<SCHEMA>.matches(object)

  Example:

  rsa_key = {'keytype': 'rsa'
             'keyid': 34892fc465ac76bc3232fab
             'keyval': {'public': 'public_key',
                        'private': 'private_key'}

  securesystemslib.formats.RSAKEY_SCHEMA.check_match(rsa_key)
  securesystemslib.formats.RSAKEY_SCHEMA.matches(rsa_key)

  In this example, if a dict key or dict value is missing or incorrect,
  the match fails.  There are numerous variations of object checking
  provided by 'formats.py' and 'schema.py'.

  The second section deals with the role metadata classes.  There are
  multiple top-level roles, each with differing metadata formats.
  Example:

  root_object = securesystemslib.formats.RootFile.from_metadata(root_metadata_file)
  targets_metadata = securesystemslib.formats.TargetsFile.make_metadata(...)

  The input and output of these classes are checked against their respective
  schema to ensure correctly formatted metadata.

  The last section contains miscellaneous functions related to the format of
  securesystemslib objects.
  Example:

  signable_object = make_signable(unsigned_object)
"""

# Help with Python 3 compatibility, where the print statement is a function, an
# implicit relative import is invalid, and the '/' operator performs true
# division.  Example:  print 'hello world' raises a 'SyntaxError' exception.
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import binascii
import calendar
import re
import datetime
import time
import six

from securesystemslib import exceptions
from securesystemslib import schema as SCHEMA

# Note that in the schema definitions below, the 'SCHEMA.Object' types allow
# additional keys which are not defined. Thus, any additions to them will be
# easily backwards compatible with clients that are already deployed.

ANY_STRING_SCHEMA = SCHEMA.AnyString()
LIST_OF_ANY_STRING_SCHEMA = SCHEMA.ListOf(ANY_STRING_SCHEMA)

# A datetime in 'YYYY-MM-DDTHH:MM:SSZ' ISO 8601 format.  The "Z" zone designator
# for the zero UTC offset is always used (i.e., a numerical offset is not
# supported.)  Example: '2015-10-21T13:20:00Z'.  Note:  This is a simple format
# check, and an ISO8601 string should be fully verified when it is parsed.
ISO8601_DATETIME_SCHEMA = SCHEMA.RegularExpression(r'\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z')

# A Unix/POSIX time format.  An integer representing the number of seconds
# since the epoch (January 1, 1970.)  Metadata uses this format for the
# 'expires' field.  Set 'hi' to the upper timestamp limit (year 2038), the max
# value of an int.
UNIX_TIMESTAMP_SCHEMA = SCHEMA.Integer(lo=0, hi=2147483647)

# A hexadecimal value in '23432df87ab..' format.
HEX_SCHEMA = SCHEMA.RegularExpression(r'[a-fA-F0-9]+')

HASH_SCHEMA = HEX_SCHEMA

# A dict in {'sha256': '23432df87ab..', 'sha512': '34324abc34df..', ...} format.
HASHDICT_SCHEMA = SCHEMA.DictOf(
  key_schema = SCHEMA.AnyString(),
  value_schema = HASH_SCHEMA)

# Uniform Resource Locator identifier (e.g., 'https://www.updateframework.com/').
# TODO: Some level of restriction here would be good....  Note that I pulled
#       this from securesystemslib, since it's neither sophisticated nor used
#       by anyone else.
URL_SCHEMA = SCHEMA.AnyString()

# A key identifier (e.g., a hexadecimal value identifying an RSA key).
KEYID_SCHEMA = HASH_SCHEMA

# A list of KEYID_SCHEMA.
KEYIDS_SCHEMA = SCHEMA.ListOf(KEYID_SCHEMA)

# The signing scheme used by a key to generate a signature (e.g.,
# 'rsassa-pss-sha256' is one of the signing schemes for key type 'rsa').
SCHEME_SCHEMA = SCHEMA.AnyString()

# A path string, whether relative or absolute, e.g. 'metadata/root/'
PATH_SCHEMA = SCHEMA.AnyNonemptyString()
PATHS_SCHEMA = SCHEMA.ListOf(PATH_SCHEMA)

# An integer representing logger levels, such as logging.CRITICAL (=50).
# Must be between 0 and 50.
LOGLEVEL_SCHEMA = SCHEMA.Integer(lo=0, hi=50)

# A string representing a named object.
NAME_SCHEMA = SCHEMA.AnyString()
NAMES_SCHEMA = SCHEMA.ListOf(NAME_SCHEMA)

# A byte string representing data.
DATA_SCHEMA = SCHEMA.AnyBytes()

# A text string.  For instance, a string entered by the user.
TEXT_SCHEMA = SCHEMA.AnyString()

# Supported hash algorithms.
HASHALGORITHMS_SCHEMA = SCHEMA.ListOf(SCHEMA.OneOf(
  [SCHEMA.String('md5'), SCHEMA.String('sha1'),
   SCHEMA.String('sha224'), SCHEMA.String('sha256'),
   SCHEMA.String('sha384'), SCHEMA.String('sha512'),
   SCHEMA.String('blake2s'), SCHEMA.String('blake2b'),
   SCHEMA.String('blake2b-256')]))

# The contents of an encrypted key.  Encrypted keys are saved to files
# in this format.
ENCRYPTEDKEY_SCHEMA = SCHEMA.AnyString()

# A value that is either True or False, on or off, etc.
BOOLEAN_SCHEMA = SCHEMA.Boolean()

# The minimum number of bits for an RSA key.  Must be 2048 bits, or greater
# (recommended by TUF).  Recommended RSA key sizes:
# http://www.emc.com/emc-plus/rsa-labs/historical/twirl-and-rsa-key-size.htm#table1
RSAKEYBITS_SCHEMA = SCHEMA.Integer(lo=2048)

# The supported ECDSA signature schemes
ECDSA_SCHEME_SCHEMA = SCHEMA.RegularExpression(r'ecdsa-sha2-nistp(256|384)')

# A pyca-cryptography signature.
PYCACRYPTOSIGNATURE_SCHEMA = SCHEMA.AnyBytes()

# An RSA key in PEM format.
PEMRSA_SCHEMA = SCHEMA.AnyString()

# An ECDSA key in PEM format.
PEMECDSA_SCHEMA = SCHEMA.AnyString()

# A string representing a password.
PASSWORD_SCHEMA = SCHEMA.AnyString()

# A list of passwords.
PASSWORDS_SCHEMA = SCHEMA.ListOf(PASSWORD_SCHEMA)

# The actual values of a key, as opposed to meta data such as a key type and
# key identifier ('rsa', 233df889cb).  For RSA keys, the key value is a pair of
# public and private keys in PEM Format stored as strings.
KEYVAL_SCHEMA = SCHEMA.Object(
  object_name = 'KEYVAL_SCHEMA',
  public = SCHEMA.AnyString(),
  private = SCHEMA.Optional(SCHEMA.AnyString()))

# Public keys CAN have a private portion (for backwards compatibility) which
# MUST be an empty string
PUBLIC_KEYVAL_SCHEMA = SCHEMA.Object(
  object_name = 'KEYVAL_SCHEMA',
  public = SCHEMA.AnyString(),
  private = SCHEMA.Optional(SCHEMA.String("")))

# Supported securesystemslib key types.
KEYTYPE_SCHEMA = SCHEMA.OneOf(
  [SCHEMA.String('rsa'), SCHEMA.String('ed25519'), SCHEMA.String('ecdsa'),
   SCHEMA.RegularExpression(r'ecdsa-sha2-nistp(256|384)')])

# A generic securesystemslib key.  All securesystemslib keys should be saved to
# metadata files in this format.
KEY_SCHEMA = SCHEMA.Object(
  object_name = 'KEY_SCHEMA',
  keytype = SCHEMA.AnyString(),
  scheme = SCHEME_SCHEMA,
  keyval = KEYVAL_SCHEMA,
  expires = SCHEMA.Optional(ISO8601_DATETIME_SCHEMA))

# Like KEY_SCHEMA, but requires keyval's private portion to be unset or empty,
# and optionally includes the supported keyid hash algorithms used to generate
# the key's keyid.
PUBLIC_KEY_SCHEMA = SCHEMA.Object(
  object_name = 'PUBLIC_KEY_SCHEMA',
  keytype = SCHEMA.AnyString(),
  keyid_hash_algorithms = SCHEMA.Optional(HASHALGORITHMS_SCHEMA),
  keyval = PUBLIC_KEYVAL_SCHEMA,
  expires = SCHEMA.Optional(ISO8601_DATETIME_SCHEMA))

# A securesystemslib key object.  This schema simplifies validation of keys
# that may be one of the supported key types.  Supported key types: 'rsa',
# 'ed25519'.
ANYKEY_SCHEMA = SCHEMA.Object(
  object_name = 'ANYKEY_SCHEMA',
  keytype = KEYTYPE_SCHEMA,
  scheme = SCHEME_SCHEMA,
  keyid = KEYID_SCHEMA,
  keyid_hash_algorithms = SCHEMA.Optional(HASHALGORITHMS_SCHEMA),
  keyval = KEYVAL_SCHEMA,
  expires = SCHEMA.Optional(ISO8601_DATETIME_SCHEMA))

# A list of securesystemslib key objects.
ANYKEYLIST_SCHEMA = SCHEMA.ListOf(ANYKEY_SCHEMA)

# RSA signature schemes.
RSA_SCHEME_SCHEMA = SCHEMA.OneOf([
  SCHEMA.RegularExpression(r'rsassa-pss-(md5|sha1|sha224|sha256|sha384|sha512)'),
  SCHEMA.RegularExpression(r'rsa-pkcs1v15-(md5|sha1|sha224|sha256|sha384|sha512)')])

# An RSA securesystemslib key.
RSAKEY_SCHEMA = SCHEMA.Object(
  object_name = 'RSAKEY_SCHEMA',
  keytype = SCHEMA.String('rsa'),
  scheme = RSA_SCHEME_SCHEMA,
  keyid = KEYID_SCHEMA,
  keyid_hash_algorithms = SCHEMA.Optional(HASHALGORITHMS_SCHEMA),
  keyval = KEYVAL_SCHEMA)

# An ECDSA securesystemslib key.
ECDSAKEY_SCHEMA = SCHEMA.Object(
  object_name = 'ECDSAKEY_SCHEMA',
  keytype = SCHEMA.OneOf([SCHEMA.String('ecdsa'),
                          SCHEMA.RegularExpression(r'ecdsa-sha2-nistp(256|384)')]),
  scheme = ECDSA_SCHEME_SCHEMA,
  keyid = KEYID_SCHEMA,
  keyid_hash_algorithms = SCHEMA.Optional(HASHALGORITHMS_SCHEMA),
  keyval = KEYVAL_SCHEMA)

# An ED25519 raw public key, which must be 32 bytes.
ED25519PUBLIC_SCHEMA = SCHEMA.LengthBytes(32)

# An ED25519 raw seed key, which must be 32 bytes.
ED25519SEED_SCHEMA = SCHEMA.LengthBytes(32)

# An ED25519 raw signature, which must be 64 bytes.
ED25519SIGNATURE_SCHEMA = SCHEMA.LengthBytes(64)

# An ECDSA signature.
ECDSASIGNATURE_SCHEMA = SCHEMA.AnyBytes()

# Ed25519 signature schemes.  The vanilla Ed25519 signature scheme is currently
# supported.
ED25519_SIG_SCHEMA = SCHEMA.OneOf([SCHEMA.String('ed25519')])

# An ed25519 key.
ED25519KEY_SCHEMA = SCHEMA.Object(
  object_name = 'ED25519KEY_SCHEMA',
  keytype = SCHEMA.String('ed25519'),
  scheme = ED25519_SIG_SCHEMA,
  keyid = KEYID_SCHEMA,
  keyid_hash_algorithms = SCHEMA.Optional(HASHALGORITHMS_SCHEMA),
  keyval = KEYVAL_SCHEMA)

# GPG key scheme definitions
GPG_HASH_ALGORITHM_STRING = "pgp+SHA2"
GPG_RSA_PUBKEY_METHOD_STRING = "pgp+rsa-pkcsv1.5"
GPG_DSA_PUBKEY_METHOD_STRING = "pgp+dsa-fips-180-2"
GPG_ED25519_PUBKEY_METHOD_STRING = "pgp+eddsa-ed25519"

def _create_gpg_pubkey_with_subkey_schema(pubkey_schema):
  """Helper method to extend the passed public key schema with an optional
  dictionary of sub public keys "subkeys" with the same schema."""
  schema = pubkey_schema
  subkey_schema_tuple =  ("subkeys", SCHEMA.Optional(
        SCHEMA.DictOf(
          key_schema=KEYID_SCHEMA,
          value_schema=pubkey_schema
          )
        )
      )
  # Any subclass of `securesystemslib.schema.Object` stores the schemas that
  # define the attributes of the object in its `_required` property, even if
  # such a schema is of type `Optional`.
  # TODO: Find a way that does not require to access a protected member
  schema._required.append(subkey_schema_tuple) # pylint: disable=protected-access
  return schema

GPG_RSA_PUBKEYVAL_SCHEMA = SCHEMA.Object(
  object_name = "GPG_RSA_PUBKEYVAL_SCHEMA",
  e = SCHEMA.AnyString(),
  n = HEX_SCHEMA
)

# We have to define GPG_RSA_PUBKEY_SCHEMA in two steps, because it is
# self-referential. Here we define a shallow _GPG_RSA_PUBKEY_SCHEMA, which we
# use below to create the self-referential GPG_RSA_PUBKEY_SCHEMA.
_GPG_RSA_PUBKEY_SCHEMA = SCHEMA.Object(
  object_name = "GPG_RSA_PUBKEY_SCHEMA",
  type = SCHEMA.String("rsa"),
  method = SCHEMA.String(GPG_RSA_PUBKEY_METHOD_STRING),
  hashes = SCHEMA.ListOf(SCHEMA.String(GPG_HASH_ALGORITHM_STRING)),
  creation_time = SCHEMA.Optional(UNIX_TIMESTAMP_SCHEMA),
  validity_period = SCHEMA.Optional(SCHEMA.Integer(lo=0)),
  keyid = KEYID_SCHEMA,
  keyval = SCHEMA.Object(
      public = GPG_RSA_PUBKEYVAL_SCHEMA,
      private = SCHEMA.String("")
    )
)
GPG_RSA_PUBKEY_SCHEMA = _create_gpg_pubkey_with_subkey_schema(
    _GPG_RSA_PUBKEY_SCHEMA)

GPG_DSA_PUBKEYVAL_SCHEMA = SCHEMA.Object(
  object_name = "GPG_DSA_PUBKEYVAL_SCHEMA",
  y = HEX_SCHEMA,
  p = HEX_SCHEMA,
  q = HEX_SCHEMA,
  g = HEX_SCHEMA
)

# C.f. comment above _GPG_RSA_PUBKEY_SCHEMA definition
_GPG_DSA_PUBKEY_SCHEMA = SCHEMA.Object(
  object_name = "GPG_DSA_PUBKEY_SCHEMA",
  type = SCHEMA.String("dsa"),
  method = SCHEMA.String(GPG_DSA_PUBKEY_METHOD_STRING),
  hashes = SCHEMA.ListOf(SCHEMA.String(GPG_HASH_ALGORITHM_STRING)),
  creation_time = SCHEMA.Optional(UNIX_TIMESTAMP_SCHEMA),
  validity_period = SCHEMA.Optional(SCHEMA.Integer(lo=0)),
  keyid = KEYID_SCHEMA,
  keyval = SCHEMA.Object(
      public = GPG_DSA_PUBKEYVAL_SCHEMA,
      private = SCHEMA.String("")
    )
)

GPG_DSA_PUBKEY_SCHEMA = _create_gpg_pubkey_with_subkey_schema(
    _GPG_DSA_PUBKEY_SCHEMA)

GPG_ED25519_PUBKEYVAL_SCHEMA = SCHEMA.Object(
  object_name = "GPG_ED25519_PUBKEYVAL_SCHEMA",
  q = HEX_SCHEMA,
)

# C.f. comment above _GPG_RSA_PUBKEY_SCHEMA definition
_GPG_ED25519_PUBKEY_SCHEMA = SCHEMA.Object(
  object_name = "GPG_ED25519_PUBKEY_SCHEMA",
  type = SCHEMA.String("eddsa"),
  method = SCHEMA.String(GPG_ED25519_PUBKEY_METHOD_STRING),
  hashes = SCHEMA.ListOf(SCHEMA.String(GPG_HASH_ALGORITHM_STRING)),
  creation_time = SCHEMA.Optional(UNIX_TIMESTAMP_SCHEMA),
  validity_period = SCHEMA.Optional(SCHEMA.Integer(lo=0)),
  keyid = KEYID_SCHEMA,
  keyval = SCHEMA.Object(
      public = GPG_ED25519_PUBKEYVAL_SCHEMA,
      private = SCHEMA.String("")
    )
)
GPG_ED25519_PUBKEY_SCHEMA = _create_gpg_pubkey_with_subkey_schema(
    _GPG_ED25519_PUBKEY_SCHEMA)

GPG_PUBKEY_SCHEMA = SCHEMA.OneOf([GPG_RSA_PUBKEY_SCHEMA,
    GPG_DSA_PUBKEY_SCHEMA, GPG_ED25519_PUBKEY_SCHEMA])

GPG_SIGNATURE_SCHEMA = SCHEMA.Object(
    object_name = "SIGNATURE_SCHEMA",
    keyid = KEYID_SCHEMA,
    short_keyid = SCHEMA.Optional(KEYID_SCHEMA),
    other_headers = HEX_SCHEMA,
    signature = HEX_SCHEMA,
    info = SCHEMA.Optional(SCHEMA.Any()),
  )

# A single signature of an object.  Indicates the signature, and the KEYID of
# the signing key.  I debated making the signature schema not contain the key
# ID and instead have the signatures of a file be a dictionary with the key
# being the keyid and the value being the signature schema without the keyid.
# That would be under the argument that a key should only be able to sign a
# file once.
SIGNATURE_SCHEMA = SCHEMA.Object(
  object_name = 'SIGNATURE_SCHEMA',
  keyid = KEYID_SCHEMA,
  sig = HEX_SCHEMA)

# A dict where the dict keys hold a keyid and the dict values a key object.
KEYDICT_SCHEMA = SCHEMA.DictOf(
  key_schema = KEYID_SCHEMA,
  value_schema = KEY_SCHEMA)

ANY_SIGNATURE_SCHEMA = SCHEMA.OneOf([SIGNATURE_SCHEMA,
    GPG_SIGNATURE_SCHEMA])

# List of ANY_SIGNATURE_SCHEMA.
SIGNATURES_SCHEMA = SCHEMA.ListOf(ANY_SIGNATURE_SCHEMA)

# A signable object.  Holds the signing role and its associated signatures.
SIGNABLE_SCHEMA = SCHEMA.Object(
  object_name = 'SIGNABLE_SCHEMA',
  signed = SCHEMA.Any(),
  signatures = SIGNATURES_SCHEMA)

# Note: Verification keys can have private portions but in case of GPG we
# only have a PUBKEY_SCHEMA (because we never export private gpg keys from
# the gpg keyring)
ANY_VERIFICATION_KEY_SCHEMA = SCHEMA.OneOf([ANYKEY_SCHEMA,
    GPG_PUBKEY_SCHEMA])

VERIFICATION_KEY_DICT_SCHEMA = SCHEMA.DictOf(
  key_schema = KEYID_SCHEMA,
  value_schema = ANY_VERIFICATION_KEY_SCHEMA)

ANY_KEYDICT_SCHEMA = SCHEMA.OneOf([KEYDICT_SCHEMA,
    VERIFICATION_KEY_DICT_SCHEMA])

ANY_PUBKEY_SCHEMA = SCHEMA.OneOf([PUBLIC_KEY_SCHEMA, GPG_PUBKEY_SCHEMA])

ANY_PUBKEY_DICT_SCHEMA = SCHEMA.DictOf(
  key_schema = KEYID_SCHEMA,
  value_schema = ANY_PUBKEY_SCHEMA)





def datetime_to_unix_timestamp(datetime_object):
  """
  <Purpose>
    Convert 'datetime_object' (in datetime.datetime()) format) to a Unix/POSIX
    timestamp.  For example, Python's time.time() returns a Unix timestamp, and
    includes the number of microseconds.  'datetime_object' is converted to UTC.

    >>> datetime_object = datetime.datetime(1985, 10, 26, 1, 22)
    >>> timestamp = datetime_to_unix_timestamp(datetime_object)
    >>> timestamp
    499137720

  <Arguments>
    datetime_object:
      The datetime.datetime() object to convert to a Unix timestamp.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'datetime_object' is not a
    datetime.datetime() object.

  <Side Effects>
    None.

  <Returns>
    A unix (posix) timestamp (e.g., 499137660).
  """

  # Is 'datetime_object' a datetime.datetime() object?
  # Raise 'securesystemslib.exceptions.FormatError' if not.
  if not isinstance(datetime_object, datetime.datetime):
    message = repr(datetime_object) + ' is not a datetime.datetime() object.'
    raise exceptions.FormatError(message)

  unix_timestamp = calendar.timegm(datetime_object.timetuple())

  return unix_timestamp




def unix_timestamp_to_datetime(unix_timestamp):
  """
  <Purpose>
    Convert 'unix_timestamp' (i.e., POSIX time, in UNIX_TIMESTAMP_SCHEMA format)
    to a datetime.datetime() object.  'unix_timestamp' is the number of seconds
    since the epoch (January 1, 1970.)

    >>> datetime_object = unix_timestamp_to_datetime(1445455680)
    >>> datetime_object
    datetime.datetime(2015, 10, 21, 19, 28)

  <Arguments>
    unix_timestamp:
      An integer representing the time (e.g., 1445455680).  Conformant to
      'securesystemslib.formats.UNIX_TIMESTAMP_SCHEMA'.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'unix_timestamp' is improperly
    formatted.

  <Side Effects>
    None.

  <Returns>
    A datetime.datetime() object corresponding to 'unix_timestamp'.
  """

  # Is 'unix_timestamp' properly formatted?
  # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
  UNIX_TIMESTAMP_SCHEMA.check_match(unix_timestamp)

  # Convert 'unix_timestamp' to a 'time.struct_time',  in UTC.  The Daylight
  # Savings Time (DST) flag is set to zero.  datetime.fromtimestamp() is not
  # used because it returns a local datetime.
  struct_time = time.gmtime(unix_timestamp)

  # Extract the (year, month, day, hour, minutes, seconds) arguments for the
  # datetime object to be returned.
  datetime_object = datetime.datetime(*struct_time[:6])

  return datetime_object




def format_base64(data):
  """
  <Purpose>
    Return the base64 encoding of 'data' with whitespace and '=' signs omitted.

  <Arguments>
    data:
      Binary or buffer of data to convert.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if the base64 encoding fails or the
    argument is invalid.

  <Side Effects>
    None.

  <Returns>
    A base64-encoded string.
  """

  try:
    return binascii.b2a_base64(data).decode('utf-8').rstrip('=\n ')

  except (TypeError, binascii.Error) as e:
    raise exceptions.FormatError('Invalid base64'
      ' encoding: ' + str(e))




def parse_base64(base64_string):
  """
  <Purpose>
    Parse a base64 encoding with whitespace and '=' signs omitted.

  <Arguments>
    base64_string:
      A string holding a base64 value.

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'base64_string' cannot be parsed
    due to an invalid base64 encoding.

  <Side Effects>
    None.

  <Returns>
    A byte string representing the parsed based64 encoding of
    'base64_string'.
  """

  if not isinstance(base64_string, six.string_types):
    message = 'Invalid argument: '+repr(base64_string)
    raise exceptions.FormatError(message)

  extra = len(base64_string) % 4
  if extra:
    padding = '=' * (4 - extra)
    base64_string = base64_string + padding

  try:
    return binascii.a2b_base64(base64_string.encode('utf-8'))

  except (TypeError, binascii.Error) as e:
    raise exceptions.FormatError('Invalid base64'
      ' encoding: ' + str(e))




def _canonical_string_encoder(string):
  """
  <Purpose>
    Encode 'string' to canonical string format.

  <Arguments>
    string:
      The string to encode.

  <Exceptions>
    None.

  <Side Effects>
    None.

  <Returns>
    A string with the canonical-encoded 'string' embedded.
  """

  string = '"%s"' % re.sub(r'(["\\])', r'\\\1', string)

  return string


def _encode_canonical(object, output_function):
  # Helper for encode_canonical.  Older versions of json.encoder don't
  # even let us replace the separators.

  if isinstance(object, six.string_types):
    output_function(_canonical_string_encoder(object))
  elif object is True:
    output_function("true")
  elif object is False:
    output_function("false")
  elif object is None:
    output_function("null")
  elif isinstance(object, six.integer_types):
    output_function(str(object))
  elif isinstance(object, (tuple, list)):
    output_function("[")
    if len(object):
      for item in object[:-1]:
        _encode_canonical(item, output_function)
        output_function(",")
      _encode_canonical(object[-1], output_function)
    output_function("]")
  elif isinstance(object, dict):
    output_function("{")
    if len(object):
      items = sorted(six.iteritems(object))
      for key, value in items[:-1]:
        output_function(_canonical_string_encoder(key))
        output_function(":")
        _encode_canonical(value, output_function)
        output_function(",")
      key, value = items[-1]
      output_function(_canonical_string_encoder(key))
      output_function(":")
      _encode_canonical(value, output_function)
    output_function("}")
  else:
    raise exceptions.FormatError('I cannot encode '+repr(object))


def encode_canonical(object, output_function=None):
  """
  <Purpose>
    Encode 'object' in canonical JSON form, as specified at
    http://wiki.laptop.org/go/Canonical_JSON .  It's a restricted
    dialect of JSON in which keys are always lexically sorted,
    there is no whitespace, floats aren't allowed, and only quote
    and backslash get escaped.  The result is encoded in UTF-8,
    and the resulting bits are passed to output_function (if provided),
    or joined into a string and returned.

    Note: This function should be called prior to computing the hash or
    signature of a JSON object in securesystemslib.  For example, generating a
    signature of a signing role object such as 'ROOT_SCHEMA' is required to
    ensure repeatable hashes are generated across different json module
    versions and platforms.  Code elsewhere is free to dump JSON objects in any
    format they wish (e.g., utilizing indentation and single quotes around
    object keys).  These objects are only required to be in "canonical JSON"
    format when their hashes or signatures are needed.

    >>> encode_canonical("")
    '""'
    >>> encode_canonical([1, 2, 3])
    '[1,2,3]'
    >>> encode_canonical([])
    '[]'
    >>> encode_canonical({"A": [99]})
    '{"A":[99]}'
    >>> encode_canonical({"x" : 3, "y" : 2})
    '{"x":3,"y":2}'

  <Arguments>
    object:
      The object to be encoded.

    output_function:
      The result will be passed as arguments to 'output_function'
      (e.g., output_function('result')).

  <Exceptions>
    securesystemslib.exceptions.FormatError, if 'object' cannot be encoded or
    'output_function' is not callable.

  <Side Effects>
    The results are fed to 'output_function()' if 'output_function' is set.

  <Returns>
    A string representing the 'object' encoded in canonical JSON form.
  """

  result = None
  # If 'output_function' is unset, treat it as
  # appending to a list.
  if output_function is None:
    result = []
    output_function = result.append

  try:
    _encode_canonical(object, output_function)

  except (TypeError, exceptions.FormatError) as e:
    message = 'Could not encode ' + repr(object) + ': ' + str(e)
    raise exceptions.FormatError(message)

  # Return the encoded 'object' as a string.
  # Note: Implies 'output_function' is None,
  # otherwise results are sent to 'output_function'.
  if result is not None:
    return ''.join(result)
